import numpy as np
import rasterio as rio

from abc import ABCMeta, abstractmethod
from random import shuffle, randint, choice
from scipy.interpolate import interp2d
from statistics import mean
from tqdm import tqdm
from typing import Dict, Tuple


class World(metaclass=ABCMeta):
    """
    Class to represent the environment or world in an agent-based model.
    """

    def __init__(self, width: int, height: int, tracking: Dict = None,
                 torus: bool = True):
        self.width = width
        self.height = height
        self.grid = self.init_grid()
        self.torus = torus
        self.agents = {}
        self.tick = 0
        self.display_layer = None
        self.tracking = tracking
        self.globals = {}

        if self.tracking:
            self.track = {agent: {param: [] for param in tracking[agent]}
                          for agent in tracking}

    def init_grid(self) -> Dict:
        """
        Creates the world grid with a layer to keep track of agents in
        each cell.
        """

        grid = {}

        for i in range(self.width):
            for j in range(self.height):
                grid[(i, j)] = {'agents': []}

        return grid

    def add_layer(self, layer_name: str, file: str = None, array=None,
                  value: int = 0, display: bool = False):
        """
        Adds a new layer to the grid. Layer can be initialized with a
        given value or can be generated from a raster file or from a
        numpy array. In the latter cases, the layer is resampled to the
        world's dimensions.
        """

        if file is not None:
            with rio.open(file) as layer:
                array = layer.read(1)
                self.interp_to_grid(array, layer_name)

        elif array is not None:
            self.interp_to_grid(array, layer_name)

        else:
            for cell in self.grid:
                self.grid[cell][layer_name] = value

        if display:
            self.display_layer = layer_name

    def interp_to_grid(self, array, layer_name):
        """
        Bilinear interpolation of an array to the world's dimensions.
        """

        height, width = array.shape
        xrange = lambda x: np.linspace(0, 1, x)
        f = interp2d(xrange(width), xrange(height), array, kind='linear')
        new_arr = f(xrange(self.width), xrange(self.height))

        for i in range(self.width):
            for j in range(self.height):
                self.grid[(i, j)][layer_name] = new_arr[self.height - 1 - j, i]

    def to_torus(self, coords: Tuple) -> Tuple:
        """
        In case world is toroidal, converts coordinates that exceed its
        limits back to the grid.
        """

        x, y = coords
        return (x % self.width, y % self.height)

    def add_agent(self, agent):
        """
        Adds a newly-created agent to the dictionary of agents and to
        the grid.
        """

        self.agents[agent._id] = agent
        self.place_on_grid(agent)

    def remove_from_grid(self, agent):
        """
        Removes an agent from the grid.
        """

        self.grid[agent.coords]['agents'].remove(agent)

    def place_on_grid(self, agent):
        """
        Places an agent on the grid's layer that keeps track of where
        agents are.
        """

        self.grid[agent.coords]['agents'].append(agent)

    def random_cell(self):
        """
        Returns the coordinates of a random grid cell.
        """

        return (randint(0, self.width - 1), randint(0, self.height - 1))

    def random_empty_cell(self):
        """
        Returns the coordinates of a random grid cell with no agents
        on it.
        """

        empty_cells = [cell for cell in self.grid
                       if not self.grid[cell]['agents']]
        return choice(empty_cells)

    def save(self):
        """
        Stores the variables to be tracked at each step of the model.
        """

        for agent in self.tracking:
            if agent == 'global':
                for param in self.tracking[agent]:
                    self.track['global'][param].append(
                        self.globals[param])

            elif agent[:5] == 'grid_':
                layer = np.reshape([self.grid[(i, j)][agent[5:]]
                                    for j in range(self.height)
                                    for i in range(self.width)],
                                   (self.height, self.width))

                for param in self.tracking[agent]:
                    if param[:6] == 'count_':
                        val = param[6:]

                        if val.isdigit():
                            val = int(val)
                        self.track[agent][param].append(
                            np.count_nonzero(layer == val))

                    elif param == 'avg':
                        self.track[agent][param].append(
                            np.average(layer))

                    elif param == 'sum':
                        self.track[agent][param].append(
                            np.sum(layer))

                    elif param == 'min':
                        self.track[agent][param].append(
                            np.min(layer))

                    elif param == 'max':
                        self.track[agent][param].append(
                            np.max(layer))

            else:
                for param in self.tracking[agent]:
                    if param == 'count':
                        self.track[agent][param].append(
                            len([self.agents[_id] for _id in self.agents
                                 if self.agents[_id].breed == agent]))

                    elif param[:4] == 'avg_':
                        self.track[agent][param].append(
                            mean([getattr(self.agents[_id], param[4:])
                                  for _id in self.agents
                                  if self.agents[_id].breed == agent] or [0]))

                    elif param[:4] == 'sum_':
                        self.track[agent][param].append(
                            sum([getattr(self.agents[_id], param[4:])
                                 for _id in self.agents
                                 if self.agents[_id].breed == agent]))

                    elif param[:4] == 'min_':
                        self.track[agent][param].append(
                            min([getattr(self.agents[_id], param[4:])
                                 for _id in self.agents
                                 if self.agents[_id].breed == agent] or [0]))

                    elif param[:4] == 'max_':
                        self.track[agent][param].append(
                            max([getattr(self.agents[_id], param[4:])
                                 for _id in self.agents
                                 if self.agents[_id].breed == agent] or [0]))

    @abstractmethod
    def setup(self):
        """
        Actions to be executed to prepare the model before it starts to
        run.
        """

        raise NotImplementedError

    def step(self):
        """
        At each step of the model, each agent performs the actions
        defined in their own step method. Agents' actions are not
        parallel, but the order of the agents is shuffled at every step
        of the model. If keeping track of variables, they are saved at
        every step.
        """

        agent_ids = list(self.agents.keys())
        shuffle(agent_ids)
        for _id in agent_ids:
            if _id in self.agents:
                self.agents[_id].step()
        if self.tracking:
            self.save()
        self.tick += 1

    def iterate(self, n_steps: int):
        """
        Runs the model for a number of steps.
        """

        for i in tqdm(range(n_steps)):
            self.step()



import inspect

from math import hypot, sin, asin, cos, radians, degrees
from abc import ABCMeta, abstractmethod
from random import randint, choice
from typing import Dict, List, Tuple, Union


class Agent(metaclass=ABCMeta):
    """
    Class to represent an agent in an agent-based model.
    """

    _id = 0
    colors = ['blue', 'brown', 'cyan', 'gray', 'green', 'magenta', 'orange',
              'pink', 'purple', 'red', 'yellow']

    def __init__(self, world, coords: Tuple = None):
        self._id = Agent._id
        Agent._id += 1

        self.world = world
        self.coords = coords or (randint(0, self.world.width - 1),
                                 randint(0, self.world.height - 1))
        self.direction = 90
        self.breed = self.__class__.__name__.lower()
        self.icon = '.'
        self.color = choice(self.colors)

        self.world.add_agent(self)

    def die(self):
        """
        Remove the agent from the world.
        """

        del self.world.agents[self._id]
        self.world.grid[self.coords]['agents'].remove(self)
        del self

    def hatch(self):
        """
        Creates an agent and initializes it with the same parameters as
        oneself.
        """

        sig = inspect.signature(self.__init__)
        filter_keys = [param.name for param in sig.parameters.values()
                       if param.kind == param.POSITIONAL_OR_KEYWORD]
        filtered_dict = {filter_key: self.__dict__[filter_key]
                         for filter_key in filter_keys}
        return self.__class__(**filtered_dict)

    def move_to(self, coords: Tuple):
        """
        Places the agent in a different cell of the world grid.
        """

        self.world.remove_from_grid(self)
        self.coords = coords
        self.world.place_on_grid(self)

    def cell_here(self, layer = None):
        """
        Returns the value of a layer in the model's grid for the cell
        where the agent is. If no layer is specified, the values of all
        layers are returned.
        """
        if layer is not None:
            return self.world.grid[self.coords][layer]
        else:
            return self.world.grid[self.coords]

    def get_distance(self, coords: Tuple) -> int:
        """
        Returns the distance (in cells) from the agent to a pair of
        coordinates.
        """

        x, y = coords
        return round(hypot((x - self.coords[0]), (y - self.coords[1])))

    def cells_in_radius(self, radius: int) -> Dict:
        """
        Returns all cells and respective attributes within a distance
        of the agent.
        """

        if self.world.torus:
            neighborhood = {self.world.to_torus((x, y)):
                            self.world.grid[self.world.to_torus((x, y))]
                            for x in range(self.coords[0] - radius,
                                           self.coords[0] + radius + 1)
                            for y in range(self.coords[1] - radius,
                                           self.coords[1] + radius + 1)
                            if self.get_distance((x, y)) <= radius}
        else:
            neighborhood = {(x, y): self.world.grid[(x, y)]
                            for x in range(self.coords[0] - radius,
                                           self.coords[0] + radius + 1)
                            for y in range(self.coords[1] - radius,
                                           self.coords[1] + radius + 1)
                            if (self.get_distance((x, y)) <= radius and
                                (x, y) in self.world.grid)}

        return neighborhood

    def empty_cells_in_radius(self, radius: int) -> Dict:
        """
        Returns all empty cells (with no agents on them) and respective
        attributes within a distance of the agent.
        """

        if self.world.torus:
            neighborhood = {self.world.to_torus((x, y)):
                            self.world.grid[self.world.to_torus((x, y))]
                            for x in range(self.coords[0] - radius,
                                           self.coords[0] + radius + 1)
                            for y in range(self.coords[1] - radius,
                                           self.coords[1] + radius + 1)
                            if (self.get_distance((x, y)) <= radius and not
                                self.world.grid[self.world.to_torus((x, y))]
                                               ['agents'])}
        else:
            neighborhood = {(x, y): self.world.grid[(x, y)]
                            for x in range(self.coords[0] - radius,
                                           self.coords[0] + radius + 1)
                            for y in range(self.coords[1] - radius,
                                           self.coords[1] + radius + 1)
                            if (self.get_distance((x, y)) <= radius and
                                (x, y) in self.world.grid and not
                                self.world.grid[(x, y)]['agents'])}

        return neighborhood

    def nearest_cell(self, cells: Union[List, Dict]) -> Tuple:
        """
        Given a list or dictionary of cells, returns the coordinates of
        the cell that is nearest to the agent.
        """

        dists = {cell: self.get_distance(cell) for cell in cells}
        return min(dists, key=dists.get)

    def agents_in_radius(self, radius: int):
        """
        Returns all agents within a distance of oneself.
        """

        neighborhood = self.cells_in_radius(radius)
        neighbors = [agent for coords in neighborhood
                     for agent in self.world.grid[coords]['agents']
                     if agent is not self]
        return neighbors

    def agents_here(self) -> List:
        """
        Returns all agents located on the same cell as oneself.
        """

        return [agent for agent in self.world.grid[self.coords]['agents']
                if agent is not self]

    def nearest_agent(self, agents: List = None):
        """
        Given a list of agents, returns the agent that is nearest to
        oneself. If no list is provided, all agents are evaluated.
        """

        if agents is None:
            agents = [self.world.agents[_id] for _id in self.world.agents]
        dists = {agent: self.get_distance(agent.coords)
                 for agent in agents if agent is not self}
        return min(dists, key=dists.get)

    def turn_right(self, angle: int = 90):
        """
        Rotates the agent's direction a number of degrees to the right.
        """

        self.direction = round((self.direction - angle) % 360)

    def turn_left(self, angle: int = 90):
        """
        Rotates the agent's direction a number of degrees to the left.
        """

        self.direction = round((self.direction + angle) % 360)

    def forward(self, n_steps: int = 1):
        """
        Moves the agent a number of cells forward in the direction it
        is currently facing.
        """

        x = round(self.coords[0] + cos(radians(self.direction)) * n_steps)
        y = round(self.coords[1] + sin(radians(self.direction)) * n_steps)

        if self.world.torus:
            self.move_to(self.world.to_torus((x, y)))
        elif (x, y) in self.world.grid:
            self.move_to((x, y))

    def face_towards(self, coords: Tuple):
        """
        Turns the agent's direction towards a given pair of coordinates.
        """

        if coords != self.coords:
            xdif = coords[0] - self.coords[0]
            ydif = coords[1] - self.coords[1]
            dist = hypot(xdif, ydif)
            angle = degrees(asin(ydif / dist))

            if xdif < 0:
                self.direction = round(180 - angle)
            else:
                self.direction = round((360 + angle) % 360)

    def random_walk(self, n_steps: int = 1):
        """
        Moves the agent one cell forward in a random direction for a
        number of times.
        """

        for i in range(n_steps):
            self.turn_right(randint(0, 360))
            self.forward()

    @abstractmethod
    def step(self):
        """
        Methods to be performed by the agent at each step of the
        simulation.
        """

        raise NotImplementedError



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quaesit",
    version="0.0.3",
    license="MIT",
    author="Jonas Gregorio",
    author_email="jonas.gregorio@gmail.com",
    description="Quick and easy simulation tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jgregoriods/quaesit.git",
    packages=setuptools.find_packages(),
    keywords=["Agent-Based Model", "Simulation"],
    install_requires=[
        "matplotlib",
        "numpy",
        "rasterio",
        "tqdm",
        "scipy",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import inspect
import tkinter as tk
import matplotlib.figure as mplfig
import matplotlib.backends.backend_tkagg as tkagg
import numpy as np

from math import ceil
from typing import Dict, List


class GUI:
    """
    Class for a simple graphical-user interface for an agent-based
    model.
    """

    def __init__(self, master, model, controls: Dict = None, plots: List = None,
                 grid_keys: Dict = None, grid_color = 'viridis'):
        self.master = master
        self.master.title(model.__name__)
        self.model = model
        self.breeds = None
        self.running = True
        self.plots = plots or []
        self.plot_axes = []
        self.grid_keys = grid_keys
        self.grid_color = grid_color

        nrow = ceil((1 + len(self.plots)) / 3)
        ncol = len(self.plots) + 1 if len(self.plots) < 2 else 3

        self.figure = mplfig.Figure(figsize=(5 * ncol, 5 * nrow),
                                    dpi=(75 // nrow))
        self.ax = self.figure.add_subplot(nrow, ncol, 1)
        self.ax.tick_params(axis='both', which='both', bottom=False,
                            labelbottom=False, left=False, labelleft=False)

        if self.plots:
            i = 2
            for plot in self.plots:
                self.plot_axes.append(self.figure.add_subplot(nrow, ncol, i))
                i += 1

        self.figure.subplots_adjust(left=0.05, bottom=0.05, right=0.95,
                                    top=0.95, wspace=0.1, hspace=0.1)

        self.canvas = tkagg.FigureCanvasTkAgg(self.figure, self.master)
        self.canvas.get_tk_widget().grid(row=2, column=2, columnspan=6 * ncol,
                                         rowspan=10 * nrow)
        self.canvas.mpl_connect('button_press_event', self.on_click)

        self.toolbar_frame = tk.Frame(self.master)
        self.toolbar_frame.grid(row=10 * nrow + 2, column=2, columnspan=5)
        self.toolbar = tkagg.NavigationToolbar2Tk(self.canvas,
                                                  self.toolbar_frame)

        self.setup_button = tk.Button(master, text='Setup', width=10,
                                      command=self.setup)
        self.setup_button.grid(row=1, column=2)

        self.step_button = tk.Button(master, text='Step', width=10,
                                     command=self.step)
        self.step_button.grid(row=1, column=3)

        self.run_button = tk.Button(master, text='Run', width=10,
                                    command=self.run)
        self.run_button.grid(row=1, column=4)

        self.stop_button = tk.Button(master, text='Stop', width=10,
                                     command=self.stop)
        self.stop_button.grid(row=1, column=5)

        self.n_steps = tk.Entry(master, width=10)
        self.n_steps.grid(row=1, column=6)

        self.iterate_button = tk.Button(master, text='Iterate', width=10,
                                        command=self.iterate)
        self.iterate_button.grid(row=1, column=7)

        self.controls = controls

        self.model_vars = {}

        if self.controls is not None:
            for control in self.controls:
                if self.controls[control]['type'] == 'scale':
                    self.model_vars[control] = tk.IntVar()
                    label = self.controls[control]['label']
                    min, max = self.controls[control]['range']

                    new_slider = tk.Scale(self.master, label=label,
                                          from_=min, to=max,
                                          orient=tk.HORIZONTAL,
                                          variable=self.model_vars[control])
                    new_slider.set(self.controls[control]['value'])

                    new_slider.grid(row=len(self.model_vars) + 1, column=1)

                elif self.controls[control]['type'] == 'check':
                    self.model_vars[control] = tk.BooleanVar()
                    label = self.controls[control]['label']

                    new_check = tk.Checkbutton(self.master, text=label,
                                               variable=self.model_vars[control])
                    new_check.grid(row=len(self.model_vars) + 1, column=1)

    def plot_model(self):
        self.ax.cla()
        self.ax.set_xlim(0, self.model.width - 1)
        self.ax.set_ylim(0, self.model.height - 1)

        if self.model.display_layer:
            base = np.reshape([self.model.grid[(i, j)]
                                              [self.model.display_layer]
                               for j in range(self.model.height)
                               for i in range(self.model.width)],
                              (self.model.height, self.model.width))

            if base.dtype.kind == 'U':
                base = np.vectorize(self.grid_keys.__getitem__)(base)

            # To avoid showing the grid with color for lowest values at setup
            # if the grid is uniform.
            base_min = np.min(base)
            if base_min == np.max(base):
                base_min -= 1

            self.ax.imshow(base, cmap=self.grid_color, vmin=base_min)

        if self.model.agents:
            breeds = {self.model.agents[_id].breed
                      for _id in self.model.agents}
            for breed in breeds:

                agents = [self.model.agents[_id] for _id in self.model.agents
                          if self.model.agents[_id].breed == breed]

                points = [agent.coords for agent in agents]
                colors = [agent.color for agent in agents]

                self.ax.scatter(*zip(*points), c=colors, s=100,
                                marker=agents[0].icon)

        if self.plot_axes:
            i = 0
            for plot in self.plot_axes:
                plot.cla()
                labels = []
                for agent in self.plots[i]['data']:
                    for param in self.plots[i]['data'][agent]:
                        if self.plots[i]['type'] == 'line':
                            plot.plot(self.model.track[agent][param])
                        elif (self.plots[i]['type'] == 'hist' and
                              self.model.track[agent][param]):
                            plot.hist(self.model.track[agent][param][-1])
                        labels.append(f'{agent} {param}')
                plot.legend(tuple(labels), loc='upper right')

                i += 1

        self.canvas.draw()

    def setup(self):
        params = {k: v.get() for k, v in self.model_vars.items()}

        tracking = None
        if self.plots:
            tracking = {}
            for plot in self.plots:
                for agent, param in plot['data'].items():
                    tracking.setdefault(agent, [])
                    tracking[agent] += param

        self.running = True
        self.stop_button.configure(text='Stop')

        if inspect.isclass(self.model):
            self.model = self.model(tracking=tracking, **params)
        else:
            self.model = self.model.__class__(tracking=tracking, **params)

        self.model.setup()
        self.plot_model()

    def step(self):
        self.model.step()
        self.plot_model()

    def run(self):
        if self.running:
            self.model.step()
            self.plot_model()
            self.master.after(1, self.run)

    def stop(self):
        if self.running:
            self.running = False
            self.stop_button.configure(text='Resume')
        else:
            self.running = True
            self.stop_button.configure(text='Stop')
            self.run()

    def iterate(self):
        self.model.iterate(int(self.n_steps.get()))
        self.plot_model()

    def on_click(self, event):
        if event.inaxes is not None:
            x, y = round(event.xdata), round(event.ydata)
            print('\n')
            print(f'Cell {x} {y}')
            print(self.model.grid[(x, y)])
            print('Agents: ')
            for agent in self.model.grid[(x, y)]['agents']:
                print(f'{agent.breed} {agent._id}')
                for prop, val in vars(agent).items():
                    print(f'{prop}: {val}')
